/* BIOSCALL  Generic function for a low level bios call, all parameters are */
/* contained in the program function call.
/* A     B     C     D     E     F     G     H                              */
/* (ah)  (al)  (bh)  (bl)  (ch)  (cl)  (dh)  (dl)                           */
/*    by Jim Mendel   13967 Marquesas way # 29   Marina del Rey, Ca 90292   */
/****************************************************************************/
#include <dos.h>
BIOSCALL(int A, int B, int C, int D, int E, int F, int G, int H)
{
union REGS REG;
REG.h.ah=A;
REG.h.al=B;
REG.h.bh=C;
REG.h.bl=D;
REG.h.ch=E;
REG.h.cl=F;
REG.h.dh=G;
REG.h.dl=H;
int86(0x10,&REG,&REG);
}

/* INTENSE  Bios call to toggle blink/intensity bit to intensity state       */
/*****************************************************************************/
/* BIOSCALL(0x10,0x03,0,0,0,0,0,0);*/ /* toggle blink/intensity to intensity */
/* A        B         C         D        E        F        G        H        */
/* (ah)     (al)      (bh)      (bl)     (ch)     (cl)     (dh)     (dl)     */
/* funct    sub-      not       blink    not      not      not      not      */
/* #        funct     used      use 1    used     used     used     used     */
/*          blink/              intensity                                    */
/*          intensity           use 0                                        */



/* CLSB  Clear screen to black thru use of scroll-up bios function           */
/*****************************************************************************/
/*   BIOSCALL(6,0,7,0,0,0,50,79); */     /* clear the screen, up to 50 lines */ 
/*   BIOSCALL(2,0,0,0,0,0,0,0);   */     /* set cursor to upper left corner  */
/* A        B         C         D        E        F        G        H        */
/* (ah)     (al)      (bh)      (bl)     (ch)     (cl)     (dh)     (dl)     */
/* funct    quan      blanked   not      y1       x1       y2       x2       */
/* #        lines     area      used                                         */
/* scroll   scrolled  & cursor                                               */
/* up       (0=all)   colors                                                 */



/* CLSW  Clear screen to white thru use of scroll-up bios function           */
/*****************************************************************************/
/*   BIOSCALL(6,0,112,0,0,0,50,79); */ /* clear the screen, up to 50 lines   */ 
/*   BIOSCALL(2,0,0,0,0,0,0,0);     */ /* set cursor to upper left corner    */
/* A        B         C         D        E        F        G        H        */
/* (ah)     (al)      (bh)      (bl)     (ch)     (cl)     (dh)     (dl)     */
/* funct    quan      blanked   not      y1       x1       y2       x2       */
/* #        lines     area      used                                         */
/* (scroll  scrolled  & cursor                                               */
/* up)      (0=all)   colors                                                 */


/* SCROLLUP  Scrolls screen up                                               */
/*****************************************************************************/
/*   int i;                       */
/*   for (i=1;i<10;i++)           *//* scroll loop - 10 lines                */
/*   BIOSCALL(6,1,75,0,0,0,24,79);*//* scroll 1 line up, red scrn, blue curs */ 
/* A        B         C         D        E        F        G        H        */
/* (ah)     (al)      (bh)      (bl)     (ch)     (cl)     (dh)     (dl)     */
/* funct    quan      blanked   not      y1       x1       y2       x2       */
/* #        lines     area      used                                         */
/* (scroll  scrolled  & cursor                                               */
/* up)      (0=all)   colors                                                 */


/* SCROLLDN  Scrolls screen down                                             */
/*****************************************************************************/
/*   int i;                        */
/*   for (i=1;i<10;i++)            *//* scroll loop - 10 lines               */
/*   BIOSCALL(7,1,75,0,0,0,24,79); *//* scroll 1 line dn, red scrn, blue curs*/ 
/* A        B         C         D        E        F        G        H        */
/* (ah)     (al)      (bh)      (bl)    (ch)      (cl)     (dh)     (dl)     */
/* funct    quan      blanked   not     y1        x1       y2       x2       */
/* #        lines     area      used                                         */
/* (scroll  scrolled  & cursor                                               */
/* down)    (0=all)   colors                                                 */


/* PRNC  Print Z char at current cursor position, uses prior color attribute */
/*****************************************************************************/
/*   BIOSCALL(10,90,0,0,0,1,0,0); */      /* prints letter Z in prior color  */ 
/* A        B         C         D        E        F         G        H       */ 
/* (ah)     (al)      (bh)      (bl)     (ch)     (cl)      (dh)     (dl)    */
/* funct    ASCII     video     not      not      qn to     not      not     */
/* # (prn   char      page #    used     used     print     used     used    */
/* char)    (dec #)                                                          */


/* PRNCC  Prints color-specified character Z at current cursor position      */
/*****************************************************************************/
/*   BIOSCALL(9,90,0,4,0,1,0,0);  */               /* prints letter Z in red */
/* A        B          C         D        E        F        G         H      */ 
/* (ah)     (al)       (bh)      (bl)     (ch)     (cl)     (dh)      (dl)   */
/* funct    ASCII      video     char     not      qn to    not       not    */
/* # (prn   char       page #    color    used     print    used      used   */
/* char &)  (dec #)                                                          */
/* color)                                                                    */


/* GOTOXY  Move cursor to specified row(y) & column(x) coordinates           */
/*****************************************************************************/
/* BIOSCALL(2,0,0,0,0,0,12,40);*/ /* move cursor center-page, x-y coordinate */ 
/*   getch();                  */ /* hold cursor at spec'd location to see   */
/* A        B         C         D        E        F         G        H       */ 
/* (ah)     (al)      (bh)      (bl)     (ch)     (cl)      (dh)     (dl)    */
/* funct    not       video     not      not      not       row      column  */
/* # (move  used      page #    used     used     used      coord    coord   */
/* cursor                                                   (y)      (x)     */
