/* CENTER  Demonstrates function that centers inputted strings              */
/****************************************************************************/
#include <stdio.h>
#include <conio.h>
#include <string.h>
char *word_wrap(char *string, size_t line_len); /* function prototype       */
void set_tab_size(size_t size);                 /* function prototype       */

void center(FILE *file, char *string, size_t width)
{
      char *line,*end_line;
      size_t spaces;
      int last = 0;
      size_t len;

      word_wrap(string,width);
      line = string;
      while (!last)
      {
            end_line = strchr(line,'\n');
            if (NULL==end_line)
            {
                  last = 1;
                  end_line = strchr(line,'\0');
            }
            len = end_line - line;
            spaces = (width - len) / 2 ;
            fprintf(file,"\n%*c%*.*s",spaces,' ',len,len,line);
            line = end_line + 1;
      }
}


int main(void)
{
      char *string = "This is a long string to see if it will be"
            " printed out correctly but I'm not sure whether it will work"
            " correctly or not.  I guess we'll see when we try it out.";

      printf("\nHere's the string centered in 50 columns.\n");
      center(stdout,string,50);
      printf("\n\nAnd here it's centered in 72 columns.\n");
      center(stdout,string,72);
}

/* W_WRAP  Word-wrap function takes array & wraps to specified line width.  */
/* Words are not split across lines. It is rather simple in that: 1) It has */
/* no clue about how to hyphenate words.  2) It makes no attempt at dealing */
/* with a singly word longer than the specified line length, and 3) It does */
/* not deal with multiple spaces new-lines, etc. ( eg. has no clue of para- */
/* graph seperation, etc. )                                                 */
/* Tabs will be considered the size specified.  Note that word_wrap() does  */
/* not actually expand tabs.  This is only to inform it of the number of    */
/* spaces the output device will expand them to, so it will know how much   */
/* they expand a line.  The only time word_wrap does anything with tabs, is */
/* if the tab size is set to zero, in which case each tab is replaced with  */
/* a single space character.  This often provides the most useful output,   */
/* since tabs will often be in the wrong places after re-formatting, and is */
/* therfore the default.                                                    */
/*
**      char *word_wrap(char *string,long line_len);
**          Does the actual word-wrapping, as described above;
**          Parameters:
**            string:     actual string to work with
**            line_len:   length of lines for output
**          Returns:        pointer to justified string.
**
**          void set_tab_size(int size);
**            Set the number of spaces that tabs will be expanded to on output
**            default tab size is zero. (each tab replaced with a space char )
**            word_wrap does not actually expand tabs.  This only lets it keep
**            track of how many spaces they take up.  If this is set to
**            zero, each tab will be replaced with a single space.
**
**  Other procedures:
**      int get_word(char *string);
**          returns the number of characters in the next word in string,
**          including leading white-space characters.
*/
#include <stddef.h>
#include <ctype.h>

enum {FALSE,TRUE};

static int tab_size = 0;                  /* size to consider tabs as */

static size_t get_word(char *string);     /* returns size of next word*/

void set_tab_size(size_t size)
{
      tab_size = size;
}

char *word_wrap(char *string, size_t line_len)
{
      size_t len,                         /* length of current word */
             current_len = 0;             /* current length of line */
      size_t start_line = 0;              /* index of beginning if line */

      while (0 != (len = get_word(&string[current_len + start_line])))
      {
            if (current_len + len < line_len)
                  current_len += len;
            else
            {
                  string[start_line+current_len] = '\n';
                  start_line += current_len + 1;
                  current_len = 0;
            }
      }
      return string;
}

static size_t get_word(char *string)
{
      register int i = 0, word_len = 0;

      if (!string[0])
            return 0;
      while (isspace(string[i]))
      {
            if ('\t' == string[i])
            {
                  if (0 == tab_size)
                        string[i] = ' ';
                  else  word_len += tab_size-1;
            }
            else if ('\n' == string[i])
                  string[i]=' ';
            word_len++;
            i++;
      }
      while (string[i] && !isspace(string[i++]))
            word_len++;
      return word_len;
}


