/* CENTERS  Arrange & display a string so that it is centered on screen     */
/****************************************************************************/

main()
{
centers(12, "This is a line to be centered");
}

/************************************************************** centers() */
/* Function that displays a string so that it is centered on the screen.

   Argument list:    int row        the row for centering
                     char *s        pointer to the string to be centered
   Return value:     int            -1 if an error occurred, 0 otherwise  */
/**************************************************************************/
#include <stdio.h>
#include <string.h>

int centers(int row, char *s)
{
   int wide, len;

   len = strlen(s);
   wide = getvcols();
   if (len > wide)
      return -1;
   cursor(row, (wide - len) >> 1);
   write(fileno(stdout), s, len);
   return 0;
}


/************************************************************* getvcols() */
/* This function determines the number of columns presently being used
   by the video display via interrupt 0x10, function 0x0f. Most compilers
   require including the dos.h header file. Several other library routines
   call this function (e.g., ereol(), erabos(), etc.)

   Argument list:    void

   Return value:     int      the number of columns currently being used
                              by the video display (40 or 80).            */
/**************************************************************************/
#include <dos.h>

int getvcols(void)
{
   int mode;
   union REGS ireg;
   ireg.h.ah = 0x0f;
   int86(0x10, &ireg, &ireg);            /* mode is returned in ireg.h.al  */
   if (ireg.h.al < 2)
      return 40;                         /* 40-column mode                 */
   if (ireg.h.al > 3 && ireg.h.al != 7)
      return 0;                          /* Return 0 for any graphics mode */
   else
      return 80;                         /* 80-column mode                 */
}


/************************************************************** cursor() */
/* This function places the cursor a row-col as given by the function
   arguments. NOTE: for speed reasons, valid row and column numbers are
   NOT checked. getvcols() could be used to check valid columns; rows
   should not exceed 25. The value 1 is subtracted from row and column
   so that the upper-left corner of the screen is 1,1 rather than 0,0.

   Argument list:    int row    the row position for the cursor
                     int col    the column position for the cursor       */
/*************************************************************************/
#include <dos.h>

cursor(int row, int col)
{
   union REGS ireg;
   ireg.h.ah = 0x02;                         /* Function number 2        */
   ireg.h.bh = getvpage();                   /* Get active video page    */
   ireg.h.dh = row - 1;
   ireg.h.dl = col - 1;
   int86(0x10, &ireg, &ireg);
}

/************************************************************ getvpage() */
/* This function gets the active video page.
   Argument list:    void
   Return value:     int        the currently-active page                */
/*************************************************************************/
#include <dos.h>

int getvpage(void)
{
   union REGS ireg;
   ireg.h.ah = 0x0f;                         /* Function 0x0f            */
   int86(0x10, &ireg, &ireg);
   return (ireg.h.bh);
}
