/*
 * CTLDLG.C
 *
 * Routines to interface the DLL custom control to
 * the Dialog Editor.
 *
 * Robin J. Leatherbarrow
 *
 * Based extensively on MUSCROLL SDK example by	Kraig Brockschmidt
 */


#include <windows.h>
#include "picbdll.h"


/*
 * Globals for use in the Style Dialog.
 */

HANDLE      hStyleMem;
LPFNSTRTOID lpStrToId;
LPFNIDTOSTR lpIdToStr;


extern HANDLE hgInst;


DWORD PASCAL FormStyleFlags(HWND, DWORD);



/*
 * HCtlInfo
 *
 * Purpose:
 *  Provides basic information about the control to the caller, which
 *  is usually the Dialog Editor or any other possible future program
 *  that edits dialog tempaltes and needs to know about controls.
 *
 *  The information is contained in a global memory object of type
 *  INFO.  This memory object becomes the responsibility of the
 *  caller.
 *
 * Parameters:
 *  none
 *
 * Return Value:
 *  HANDLE      Handle to a Global memory object or NULL if it cannot
 *              be allocated.  The caller must free the memory.
 *              ALLOCATE WITH GMEM_DDESHARE!
 *
 */

HANDLE FAR __export	PASCAL HCtlInfo(void)
    {
    HANDLE      hMem;
    LPCTLINFO   pCtlInfo;

    //Allocate a INFO struct
    hMem=GlobalAlloc(GMEM_MOVEABLE, sizeof(CTLINFO));

    if (hMem==NULL)
        return NULL;

    //Get the pointers we need.
    pCtlInfo=(LPCTLINFO)GlobalLock(hMem);

    if (pCtlInfo==NULL)
        {
        GlobalFree(hMem);
        return NULL;
        }

    //Set the overall control info.
    pCtlInfo->wVersion=0100;
	pCtlInfo->wCtlTypes=1;
    LoadString(hgInst, IDS_CLASSNAME, pCtlInfo->szClass, CTLCLASS);
    LoadString(hgInst, IDS_FULLNAME,  pCtlInfo->szTitle, CTLTITLE);

    //Set the types
    pCtlInfo->Type[0].wType  =0;
	pCtlInfo->Type[0].wWidth =TOOLBUTTON_STDWIDTH;
	pCtlInfo->Type[0].wHeight=TOOLBUTTON_STDHEIGHT;
	pCtlInfo->Type[0].dwStyle=WS_CHILD | WS_VISIBLE | PBS_COMMAND | PBS_MENUSELECT;
	LoadString(hgInst, IDS_DESCR, pCtlInfo->Type[0].szDescr, CTLDESCR);


    //Give the memory to the Dialog Editor.
    GlobalUnlock(hMem);
    return hMem;
    }




/*
 * WTranslateCtlFlags
 *
 * Purpose:
 *  Translates the class style flags into a text string to be used in the
 *  output of the Dialog Editor in a CONTROL statement of a dialog script
 *  (i.e. SS_ICON or BS_OWNERDRAW).  The same text as used in the
 *  incldue file should be used here.
 *
 *  Null-terminate the string for safety since we don't know what the
 *  Dialog Editor might do.
 *
 * Parameters:
 *  dwStyle     DWORD containing the style bits.
 *  psz         LPSTR pointer to string to receive the style text.
 *  cchMax      WORD maximum length of the string.
 *
 * Return Value:
 *  WORD        Number of characters filled in the buffer.  0 on an error.
 *
 */

WORD FAR PASCAL __export WTranslateCtlFlags(DWORD dwStyle, LPSTR psz, WORD cchMax)
    {
    WORD        cch=0;
    LPSTR       pszT;

    pszT=psz;

    //Use one of these two exclusive styles.
	if (dwStyle & PBS_COMMAND)
		cch=LoadString(hgInst, IDS_PBS_COMMAND, pszT, cchMax);
    else
		cch=LoadString(hgInst, IDS_PBS_ATTRIBUTE, pszT, cchMax);

    //Adjust remaining number of characters we can store.
    cchMax-=cch;
    pszT+=cch;


	//Any of these stings can be concatenated.
	if (PBS_MENUSELECT & dwStyle)
        {
		cch=LoadString(hgInst, IDS_PBS_MENUSELECT, pszT, cchMax);
        pszT+=cch;
        cchMax-=cch;
        }

	if (PBS_AUTO & dwStyle)
        {
		cch=LoadString(hgInst, IDS_PBS_AUTO, pszT, cchMax);
        pszT+=cch;
        cchMax-=cch;
        }


    /*
     * Chop off the string 2 characters from the end since we have an extra
     * '| 'there.
     */
    cch=lstrlen(psz)-2;
    *(psz+cch)=0;

    return cch;
    }




/*
 * FShowCtlStyleDlg
 *
 * Purpose:
 *  Requests the control library to display a dialog box that
 *  allows editing of the control's styles.  The fields in a SYTLE
 *  structure must be set according to what the window is and what
 *  it contains.
 *
 * Parameters:
 *  hWnd        HWND to the control.
 *  hMem        HANDLE to global memory of the STYLE structure.
 *  lpfnStrToId LPFNSTRTOID pointer to Dialog Editor function that
 *              returns the include file symbol for a given ID.
 *  lpfnIdToStr LPFNIDTOSTR pointer to Dialog Edit function that
 *              returns the ID of a given symbol.
 *
 * Return Value:
 *  BOOL        TRUE if the STYLE structure was changed, FALSE
 *              otherwise.
 */

BOOL FAR PASCAL __export FShowCtlStyleDlg(HWND hWnd, HANDLE hMem,
										  LPFNSTRTOID lpfnStrToId,
										  LPFNIDTOSTR lpfnIdToStr)
    {
    FARPROC     lpStyleDlgProc;
    BOOL        bReturn;

    //Save the vital information for the dialog.
    hStyleMem=hMem;
    lpStrToId=lpfnStrToId;
    lpIdToStr=lpfnIdToStr;

    //Show dialog box
    lpStyleDlgProc=MakeProcInstance((FARPROC)FCtlStyleDlgProc, hgInst);

    bReturn=DialogBox(hgInst, MAKEINTRESOURCE(IDD_STYLEDIALOG),
                      hWnd, lpStyleDlgProc);

    FreeProcInstance(lpStyleDlgProc);

    return bReturn;
    }


BOOL FAR PASCAL __export HelpDlgProc(HWND hDlg, WORD wMessage, WORD wParam, LONG lParam)
    {

    switch (wMessage)
        {
		case WM_COMMAND:
            switch (wParam)
                {

				case IDOK:
					EndDialog(hDlg, TRUE);
					break;

				}
            break;

        default:
            return FALSE;
        }

    return TRUE;
    }


void DoHelp(HWND hWnd)
{
	FARPROC 	lpProc;

	//Show help dialog box
	lpProc=MakeProcInstance((FARPROC)HelpDlgProc, hgInst);

	DialogBox(hgInst, MAKEINTRESOURCE(IDD_HELPDIALOG),
			  hWnd, lpProc);

	FreeProcInstance(lpProc);

}



/*
 * FCtlStyleDlgProc
 *
 * Purpose:
 *  The dialog function for the dialog box displayed when the XXXXStyle
 *  function is called.  This is a normal dialog box proc in all respects.
 *
 * Parameters:
 *  standard
 *
 * Return Value:
 *  BOOL        Standard for a dialog box.
 */

BOOL FAR PASCAL __export FCtlStyleDlgProc(HWND hDlg, WORD wMessage, WORD wParam, LONG lParam)
    {
    LPCTLSTYLE      lpCtlStyle;
    char            szTemp[32];
    DWORD           dwID;
	WORD			wStyle;
	WORD			select;

    switch (wMessage)
        {
        case WM_INITDIALOG:
            lpCtlStyle=(LPCTLSTYLE)GlobalLock(hStyleMem);

            if (lpCtlStyle==NULL)
                return FALSE;

            (*lpIdToStr)(lpCtlStyle->wId, (LPSTR)szTemp, 32);
            SetDlgItemText(hDlg, ID_IDEDIT, szTemp);
            SetDlgItemText(hDlg, ID_TEXTEDIT, lpCtlStyle->szTitle);

            wStyle=LOWORD(lpCtlStyle->dwStyle);

			if (wStyle & PBS_COMMAND)
				select = ID_COMMAND;
			else {
				if (wStyle & PBS_AUTO)
					select = ID_RADIO;
				else
					select = ID_CHECK;
				}
			CheckRadioButton(hDlg, ID_COMMAND,
							 ID_RADIO,
							 select);


			if (lpCtlStyle->dwStyle & WS_VISIBLE)
				CheckDlgButton(hDlg, ID_VISIBLE, 1);

			if (lpCtlStyle->dwStyle & WS_DISABLED)
				CheckDlgButton(hDlg, ID_DISABLED, 1);

			if (lpCtlStyle->dwStyle & WS_TABSTOP)
				CheckDlgButton(hDlg, ID_TAB, 1);

			if (lpCtlStyle->dwStyle & WS_GROUP)
				CheckDlgButton(hDlg, ID_GROUP, 1);

			if (lpCtlStyle->dwStyle & PBS_AUTO)
				CheckDlgButton(hDlg, ID_AUTO, 1);

			if (lpCtlStyle->dwStyle & PBS_MENUSELECT)
				CheckDlgButton(hDlg, ID_MENUSELECT, 1);


            GlobalUnlock(hStyleMem);
            break;


        case WM_COMMAND:
            switch (wParam)
                {
				case IDOK:

                    //Make sure the identifier works.
                    GetDlgItemText(hDlg, ID_IDEDIT, szTemp, 32);
                    dwID=(*lpStrToId)(szTemp);


                    if (LOWORD(dwID))
                        {
                        lpCtlStyle=(LPCTLSTYLE)GlobalLock(hStyleMem);

                        if (lpCtlStyle!=NULL)
                            {
                            //Save the Style, ID, and text.
							lpCtlStyle->dwStyle=FormStyleFlags(hDlg, lpCtlStyle->dwStyle);

                            lpCtlStyle->wId=HIWORD(dwID);
                            GetDlgItemText(hDlg, ID_TEXTEDIT, lpCtlStyle->szTitle, CTLTITLE);

                            GlobalUnlock(hStyleMem);
                            EndDialog(hDlg, TRUE);
                            }
                        else
                            EndDialog(hDlg, FALSE);
                        }

                    break;


                case IDCANCEL:
                    EndDialog(hDlg, FALSE);
                    break;


				case ID_HELP:
					DoHelp(hDlg);
					break;

                default:
                    break;
                }
            break;

        default:
            return FALSE;
        }

    return TRUE;
    }







/*
 * FormStyleFlags
 *
 * Purpose:
 *	Returns a DWORD with flags set to whatever options are checked in
 *  the dialog box.  It is up to the implementer of the dialog box to
 *  code the exact bit fields in this function.
 *
 * Parameters:
 *  hDlg        HWND of the style dialog.
 *
 * Return Value:
 *	DWORD		 Style Flags value.
 */

DWORD PASCAL FormStyleFlags(HWND hDlg, DWORD flags)
    {
	WORD	wStyle=0;


	// clear the style bits
	flags &= ~(WS_VISIBLE | WS_DISABLED | WS_TABSTOP | WS_GROUP);


	if (IsDlgButtonChecked(hDlg, ID_VISIBLE))
		flags |= WS_VISIBLE;

	if (IsDlgButtonChecked(hDlg, ID_DISABLED))
		flags |= WS_DISABLED;

	if (IsDlgButtonChecked(hDlg, ID_TAB))
		flags |= WS_TABSTOP;

	if (IsDlgButtonChecked(hDlg, ID_GROUP))
		flags |= WS_GROUP;

	if (IsDlgButtonChecked(hDlg, ID_RADIO))
		wStyle |= PBS_AUTO;

	if (IsDlgButtonChecked(hDlg, ID_MENUSELECT))
		wStyle |= PBS_MENUSELECT;



	if (IsDlgButtonChecked(hDlg, ID_COMMAND))
		wStyle |= PBS_COMMAND;


	return MAKELONG(wStyle, HIWORD(flags));
    }
