
/*
 * INIT.C
 *
 * LibMain entry point and initialization code for the MicroScroll
 * custom control DLL that is likely to be used once or very
 * infrequently.
 *
 * Version 1.1, October 1991, Kraig Brockschmidt
 */


#include <windows.h>
#include <stdlib.h>   // atoi

#include "picbdll.h"


HANDLE	 hgInst;	 //Need this for dialog editor style box.



/*
 * LibMain
 *
 * Purpose:
 *  DLL-specific entry point called from LibEntry.  Initializes
 *  the DLL's heap and registers the MicroScroll custom control
 *  class.
 *
 * Parameters:
 *  hInstance       HANDLE instance of the DLL.
 *  wDataSeg        WORD segment selector of the DLL's data segment.
 *  wHeapSize       WORD byte count of the heap.
 *  lpCmdLine       LPSTR to command line used to start the module.
 *
 * Return Value:
 *  HANDLE          Instance handle of the DLL.
 *
 */

HANDLE FAR PASCAL LibMain(HANDLE hInstance, WORD wDataSeg,
                          WORD cbHeapSize, LPSTR lpCmdLine)
    {
    //Go register the control
    if (FRegisterControl(hInstance))
        {
        hgInst=hInstance;

        if (0!=cbHeapSize)
            UnlockData(0);

        return hInstance;
        }

    hgInst=NULL;
    return (HANDLE)NULL;
    }





/*
 * FRegisterControl
 *
 * Purpose:
 *  Registers the MicroScroll control class, including CS_GLOBALCLASS
 *  to make the control available to all applications in the system.
 *
 * Parameters:
 *  hInstance       HANDLE Instance of the application or DLL that will
 *                  own this class.
 *
 * Return Value:
 *  BOOL            TRUE if the class is registered, FALSE otherwise.
 *                  TRUE is also returned if the class was already
 *                  registered.
 */

BOOL PASCAL FRegisterControl(HANDLE hInstance)
    {
    static BOOL     fRegistered=FALSE;
    WNDCLASS        wc;
    char            szClass[40];

    if (!fRegistered)
        {
        //Load the class name
        if (0==LoadString(hInstance, IDS_CLASSNAME, szClass, 40))
            return FALSE;

		wc.lpfnWndProc	 = PicButWndProc;
        wc.cbClsExtra    = CBCLASSEXTRA;
        wc.cbWndExtra    = CBWINDOWEXTRA;
        wc.hInstance     = hInstance;
        wc.hIcon         = NULL;
        wc.hCursor       = LoadCursor(NULL, IDC_ARROW);
		wc.hbrBackground = NULL;
        wc.lpszMenuName  = NULL;
        wc.lpszClassName = szClass;
		wc.style		 = CS_GLOBALCLASS | CS_DBLCLKS |
                           CS_VREDRAW | CS_HREDRAW;

		fRegistered=RegisterClass(&wc);

		}

    return fRegistered;
    }



void FAR PASCAL ResetImage( HWND hWnd, PPICBUT pPB, LPSTR lpszName )
{
	int 			width, height;
	BITMAP			bitmap;
	int 			newwidth, newheight;
	RECT			rect;
	static char		buffer[100];  // ensure that this is DS: referenced
								  // so declare it static (for atoi)

	GetClientRect( hWnd, &rect );

	// initialise defaults
	//
	pPB->bLimitSize = TRUE;

	pPB->bDeleteBitmap = TRUE;

	pPB->iImage = 0;

	// for strings #nnn we load the pre-defined bitmaps
	if (lpszName[0] == '#') {
		lstrcpy( buffer, &lpszName[1] );
		pPB->iImage = atoi( buffer );
		pPB->hBit = NULL;

		// we can't be an attribute button, at least for up to TOOLIMAGE_MAX
		// (this is defined by the BTTNCUR.DLL)
		//
		if (pPB->iImage <= TOOLIMAGE_MAX)
			pPB->dwStyle |= PBS_COMMAND;
		}
	else {
		// find it in the instance of our creator
		pPB->hBit = LoadBitmap( GetWindowWord(hWnd, GWW_HINSTANCE), lpszName );
		if (!pPB->hBit) {

			// can't find it, so load a default bitmap
			// this will be used for the dialog editor, for example
			//
			pPB->hBit = LoadBitmap( hgInst, "Default" );
			pPB->bLimitSize = FALSE;
			}
		}


	if (pPB->dwStyle & PBS_COMMAND) {
		if (pPB->dwStyle & WS_DISABLED)
			pPB->uState = COMMANDBUTTON_DISABLED;
		else
			pPB->uState = COMMANDBUTTON_UP;
		}
	else {
		if (pPB->dwStyle & WS_DISABLED)
			pPB->uState = ATTRIBUTEBUTTON_DISABLED;
		else
			pPB->uState = ATTRIBUTEBUTTON_UP;
		}


	// how big (minimum size) should the button be?
	width = TOOLBUTTON_STDIMAGEWIDTH;
	height = TOOLBUTTON_STDIMAGEHEIGHT;

	if (pPB->hBit) {
		GetObject( pPB->hBit, sizeof(BITMAP), &bitmap );
		width = bitmap.bmWidth;
		height = bitmap.bmHeight;
		}

	pPB->width = width;
	pPB->height = height;


	//increase space for surround
	width += TOOLBUTTON_EXTRAWIDTH;
	height += TOOLBUTTON_EXTRAHEIGHT;


	newwidth = max(width, rect.right);
	newheight = max(height, rect.bottom);

	// if we aren't bit enough, then expand
	if (pPB->bLimitSize) {
		if ((width > rect.right) || (height > rect.bottom)) {

			pPB->bIgnoreSize = TRUE;

			SetWindowPos(hWnd, NULL, 0, 0,
						 newwidth,
						 newheight,
						 SWP_NOZORDER | SWP_NOMOVE);


			pPB->bIgnoreSize = FALSE;
			}
		}

}






/*
 * LPicButCreate
 *
 * Purpose:
 *  Handles both WM_NCCREATE and WM_CREATE messages:
 *	  WM_NCCREATE:	Allocate control PICBUT structure.	Any sort
 *                  of initialization that should force the create
 *                  to fail should be included in this case.
 *
 *	  WM_CREATE:	Fill the PICBUT structure; parse the text
 *                  for initial min, max, and position for the
 *                  MSS_TEXTHASRANGE style; clear all states;
 *                  set the expansion handle to NULL.
 *
 * Parameters:
 *  hWnd            HWND handle to the control window.
 *  iMsg            WORD message, either WM_NCCREATE or WM_CREATE.
 *	pPB				PPICBUT pointer to existing PICBUT structure.
 *
 * Return Value:
 *  LONG            0L if memory could not be allocated on WM_NCCREATE.
 *                  1L in all other cases.
 */

LONG PASCAL LPicButCreate(HWND hWnd, WORD iMsg, PPICBUT pPB,
						  LPCREATESTRUCT lpCreate)
{
	HANDLE			hMem;
	RECT			rect;


	if (WM_NCCREATE==iMsg) {
        /*
		 * Allocate extra FIXED memory for the control's PICBUT
         * structure and store that handle in the minimal window
         * extra bytes allocated for the control.
         *
         * Note that LPTR includes LMEM_ZEROINIT, so all fields are
         * initially zero.
         */
		hMem=LocalAlloc(LPTR, CBPICBUT);

        if (NULL==hMem)
            return 0L;

		SetWindowWord(hWnd, GWW_PICBUTHMEM, (HANDLE)hMem);
        }

	if (WM_CREATE==iMsg) {

        //Copy styles
		pPB->dwStyle	=lpCreate->style;

		ResetImage( hWnd, pPB, (LPSTR)lpCreate->lpszName );

		GetClientRect( hWnd, &rect );

		}

    return 1L;
}
