/*
 * MSAPI.C
 *
 * Contains the API implementation of the Picture button custom
 * control
 *
 * Robin Leatherbarrow
 */


#include <windows.h>
#include "picbdll.h"


/*
 * LPicButAPI
 *
 * Purpose:
 *  Processes any control-specific function messages for the
 *	Picture Button control.
 *
 * Parameters:
 *  hWnd            HWND handle to the control window.
 *  iMsg            WORD message identifier.
 *  wParam          WORD parameter of the message.
 *  lParam          LONG parameter of the message.
 *	pPB			PPICBUT pointer to control-specific data.
 *
 * Return Value:
 *  LONG            Varies with the message.
 *
 */



LONG PASCAL LPicButAPI(HWND hWnd, WORD iMsg, WORD wParam,
					   LONG lParam, PPICBUT pPB)
{
	int 		state;
	LPPBM_IMAGE lpImage;
	int 		width, height;
	int 		newwidth, newheight;
	RECT		rect;

	switch (iMsg) {

		// use the normal button interface to communicate

		case BM_GETCHECK:
			switch (LOBYTE(pPB->uState)) {
				case ATTRIBUTEBUTTON_INDETERMINATE:
					return 2;
				case ATTRIBUTEBUTTON_DOWN:
					return 1;
				default:
					return 0;
				}
			break;

		case BM_GETSTATE:
			switch (LOBYTE(pPB->uState)) {
				case ATTRIBUTEBUTTON_INDETERMINATE:
					state = 2;
					break;
				case ATTRIBUTEBUTTON_DOWN:
					state = 1;
					break;
				default:
					state = 0;
					break;
				}
			if (pPB->uState & BUTTONGROUP_DOWN)
				state |= 4;
			return MAKELONG(state,0);
			break;

		case BM_SETCHECK:
			// are we a command button?
			// if so we cannot check the button
			if (pPB->dwStyle & PBS_COMMAND)
				break;


			// need to preserve the hi byte
			StateClear(pPB, BUTTONGROUP_DOWN | BUTTONGROUP_LIGHTFACE );


			switch (wParam) {
				case 0:   // unchecked
					break;
				case 1:   // checked
					StateSet( pPB, BUTTONGROUP_LIGHTFACE | BUTTONGROUP_DOWN );
					break;
				case 2:   // indeterminate
					StateSet( pPB, BUTTONGROUP_LIGHTFACE );
					break;
				}

			InvalidateRect( hWnd, NULL, TRUE );

			break;

		case BM_SETSTATE:

			// need to preserve the hi byte
			StateClear(pPB, BUTTONGROUP_DOWN );

			if (wParam)
				StateSet( pPB, BUTTONGROUP_DOWN );

			InvalidateRect( hWnd, NULL, TRUE );

			break;


		// the one specific message.
		// sets the bitmap
		case PBM_SETIMAGE:

			// free the old bitmap
			if (pPB->hBit)
				if (pPB->bDeleteBitmap)
					DeleteObject( pPB->hBit );

			lpImage 	= (LPPBM_IMAGE)lParam;
			pPB->hBit	= lpImage->hBit;
			pPB->width	= lpImage->width;
			pPB->height = lpImage->height;
			pPB->iImage = lpImage->iImage;
			pPB->bLimitSize = TRUE;

			// bitmaps sent to us direct should not be deleted.
			//
			pPB->bDeleteBitmap = FALSE;

			// how big (minimum size) should the button be?
			width = pPB->width + TOOLBUTTON_EXTRAWIDTH;
			height = pPB->height + TOOLBUTTON_EXTRAHEIGHT;


			GetClientRect( hWnd, &rect );

			newwidth = max(width, rect.right);
			newheight = max(height, rect.bottom);

			// if we aren't bit enough, then expand
			if ((width > rect.right) || (height > rect.bottom)) {

				pPB->bIgnoreSize = TRUE;

				SetWindowPos(hWnd, NULL, 0, 0,
							 newwidth,
							 newheight,
							 SWP_NOZORDER | SWP_NOMOVE);


				pPB->bIgnoreSize = FALSE;
				}

			InvalidateRect( hWnd, NULL, TRUE );
			break;


		}
	return 0L;
}
