/*
 * PAINT.C
 *
 * Code to handle repainting, and dealing with button up
 *
 * Robin J. Leatherbarrow 1993
 */


#include <windows.h>
#include "picbdll.h"


// Paint the bitmap using the BTTNCUR.DLL UIToolButtonDraw
// function
//
LONG FAR PASCAL PicButPaint( HWND hWnd, PPICBUT pPB )
{
	RECT		rect;
	PAINTSTRUCT ps;
	HDC 		hDC;


	hDC = BeginPaint( hWnd, &ps );

	GetClientRect( hWnd, &rect );

	UIToolButtonDraw( hDC,
					  0,0,
					  rect.right, rect.bottom,
					  pPB->hBit, pPB->width, pPB->height,
					  pPB->iImage,
					  pPB->uState );

	EndPaint( hWnd, &ps );

	return 0L;
}




// Handle the WM_LBUTTONUP message
//
void FAR ProcessButtonup( HWND hWnd, LPARAM lParam, PPICBUT pPB )
{
	POINT		pt;
	HWND		hWndTest, hWndPrev;
	RECT		rect;

	// only valid if we have captured the mouse
	//
	if (GetCapture()!=hWnd)
		return;

	ReleaseCapture();

	// we are no longer selected.

	// do we inform with a WM_MENUSELECT command?
	//
	if (pPB->dwStyle & PBS_MENUSELECT)
		SendMessage( GetParent(hWnd), WM_MENUSELECT,
					 0,
					 MAKELONG(-1,0) );


	// are we over the control?
	//
	pt=MAKEPOINT(lParam);
	GetClientRect( hWnd, &rect );
	if (!PtInRect(&rect, pt)) {
		// not over the control
		// ensure that all is as we left it and quit

		if (pPB->uState != pPB->uOldState) {
			pPB->uState = pPB->uOldState;
			InvalidateRect(hWnd, NULL, TRUE);
			UpdateWindow(hWnd);
			}

		return;
		}


	// restore the old state
	pPB->uState = pPB->uOldState;


	// if we are a radio button, we don't flip state
	// if we are selected again
	if ((pPB->dwStyle & PBS_AUTO) &&
		   (pPB->uState == ATTRIBUTEBUTTON_DOWN)) {

		InvalidateRect( hWnd, NULL, TRUE );
		UpdateWindow( hWnd );
		return; // all done
		}

	if (pPB->dwStyle & PBS_COMMAND) {
		   }
	else {
		// flip the state. Down ----------> Up
		//				   Up ------------> Down
		//				   Indeterminate -> Down
		if (pPB->uState == ATTRIBUTEBUTTON_DOWN)
			pPB->uState = ATTRIBUTEBUTTON_UP;
		else
			pPB->uState = ATTRIBUTEBUTTON_DOWN;
		}

	// inform mummy and daddy that we have been clicked

	PostMessage( GetParent(hWnd), WM_COMMAND,
					GetWindowWord(hWnd, GWW_ID),
					MAKELONG(hWnd, BN_CLICKED) );

	// do we automatically uncheck all in our group?
	if (pPB->dwStyle & ID_AUTO) {
		   // find the start of our group
		   hWndTest = hWnd;
		   for (;;) {
			   if (GetWindowLong(hWndTest, GWL_STYLE) & WS_GROUP)
				   break;
			   hWndPrev = GetWindow( hWndTest, GW_HWNDPREV );
			   if (!hWndPrev)
				   break;
			   hWndTest = hWndPrev;
			   }
		   // hWndTest holds the start
		   while (hWndTest) {
			   if (hWndTest != hWnd)
				   SendMessage( hWndTest, BM_SETCHECK, 0, 0L );
			   hWndTest = GetWindow( hWndTest, GW_HWNDNEXT );
			   if (hWndTest)
				   if (GetWindowLong(hWndTest, GWL_STYLE) & WS_GROUP)
					   break;
			   }

		   }

	InvalidateRect( hWnd, NULL, TRUE );
	UpdateWindow( hWnd );

}
