/*
 * PICBDLL.H
 *
 * DLL Specific include file.
 *
 * Contains all definitions and and prototypes pertinent ONLY
 * to the DLL.	API related information is contained in PICBUT.H.
 *
 * Robin Leatherbarrow
 *
 */


#include <custcntl.h>   //Standard Windows header file for custom controls.
#include "picbut.h"		//Get interface stuff for the control.
#include <bttncur.h>	//MS Toolbar drawing DLL


/*
 * In window extra bytes we simply store a local handle to
 * a PICBUT data structure.  The local memory is allocated
 * from the control's local heap (either application or DLL)
 * instead of from USER's heap, thereby saving system resources.
 *
 * Note that the window styles that are stored in the regular
 * windwow structure are copied here.  This is to optimize access
 * to these bits, avoiding extra calls to GetWindowLong.
 */

typedef struct tagPICBUT
    {
    DWORD       dwStyle;        //Copy of GetWindowLong(hWnd, GWL_STYLE)
	UINT		uState;			//State flags
	UINT		uOldState;		//copy of state flags, so we can restore the old state
	HGLOBAL 	hBit;			//handle of loaded bitmap resource
	int 		width;			//width of bitmap
	int 		height; 		//height of bitmap
	BOOL		bLimitSize; 	//do we restrict the size of the button?
	int 		iImage; 		//Offset of image in the bitmap
	BOOL		bIgnoreSize;	//internal use. Set to TRUE to ignore WM_SIZE
	BOOL		bDeleteBitmap;	//internal use. Should we delete the bitmap after use?
	} PICBUT;

typedef PICBUT	   *PPICBUT;
typedef PICBUT FAR *LPPICBUT;

#define CBPICBUT sizeof(PICBUT)



//Offsets to use with GetWindowWord
#define GWW_PICBUTHMEM	0

//Extra bytes for the window if the size of a local handle.
#define CBWINDOWEXTRA       sizeof(HANDLE)

//Extra Class bytes.
#define CBCLASSEXTRA        0



/*
 * Macros to change the control state given a PPICBUT
 * and state flag(s)
 */
#define StateSet(p, wFlags)    (p->uState |=  (wFlags))
#define StateClear(p, wFlags)  (p->uState &= ~(wFlags))
#define StateTest(p, wFlags)   (p->uState &	  (wFlags))




//Private functions specific to the control.

//INIT.C
HANDLE FAR PASCAL LibMain(HANDLE, WORD, WORD, LPSTR);
BOOL       PASCAL FRegisterControl(HANDLE);
LONG	   PASCAL LPicButCreate(HWND, WORD, PPICBUT, LPCREATESTRUCT);
void   FAR PASCAL ResetImage( HWND hWnd, PPICBUT pPB, LPSTR lpszName );


//MSAPI.C
LONG	   PASCAL LPicButAPI(HWND, WORD, WORD, LONG, PPICBUT);



//PAINT.C
LONG FAR PASCAL PicButPaint( HWND, PPICBUT	);
void FAR ProcessButtonup( HWND, LPARAM, PPICBUT );


//WEP.C
void   FAR PASCAL WEP(int);


//PICBUT.C
LONG FAR PASCAL __export PicButWndProc(HWND, UINT, WPARAM, LPARAM);



//Stringtable identifiers
#define IDS_CLASSNAME           0
#define IDS_FULLNAME            1
#define IDS_CREDITS             2

//Dialog editor interface string
#define IDS_DESCR				3

#define IDS_PBS_COMMAND 		4
#define IDS_PBS_ATTRIBUTE		5

#define IDS_PBS_MENUSELECT		7
#define IDS_PBS_AUTO			8




#define TOOLBUTTON_EXTRAWIDTH (TOOLBUTTON_STDWIDTH - TOOLBUTTON_STDIMAGEWIDTH)
#define TOOLBUTTON_EXTRAHEIGHT (TOOLBUTTON_STDHEIGHT - TOOLBUTTON_STDIMAGEHEIGHT)


/*
 * All that follows in this include file is specific to
 * the dialog editor interface for the control.
 */

#include "picb_dlg.h"





//Standard dialog editor callbacks.
HANDLE FAR PASCAL __export HCtlInfo(void);
WORD   FAR PASCAL __export WTranslateCtlFlags(DWORD, LPSTR, WORD);
BOOL   FAR PASCAL __export FShowCtlStyleDlg(HWND, HANDLE, LPFNSTRTOID, LPFNIDTOSTR);
BOOL   FAR PASCAL __export FCtlStyleDlgProc(HWND, WORD, WORD, LONG);
