/*
 * MUSCROLL.C
 *
 * Contains the main window procedure of the MicroScroll control
 * that handles mouse logic, and Windows messages.
 *
 * Version 1.1, October 1991, Kraig Brockschmidt
 */

#include <windows.h>
#include "picbdll.h"
#include "picbtest.h"





/*
 * WinMain
 *
 * Purpose:
 *  Main entry point of application.   Should register the app class
 *  if a previous instance has not done so and do any other one-time
 *  initializations.
 *
 * Parameters:
 *  Standard
 *
 * Return Value:
 *  int              Value to return to Windows--termination code.
 *
 */

int PASCAL WinMain (HANDLE hInstance, HANDLE hPrevInstance,
                    LPSTR lpszCmdLine, int nCmdShow)
    {
    WNDCLASS    wc;
    HWND        hWnd;
	MSG 		msg;
	HANDLE		hLib;


    if (!hPrevInstance)
        {
        wc.style          = CS_HREDRAW | CS_VREDRAW;
		wc.lpfnWndProc	  = TestWndProc;
        wc.cbClsExtra     = 0;
		wc.cbWndExtra	  = DLGWINDOWEXTRA;
        wc.hInstance      = hInstance;
		wc.hIcon		  = LoadIcon(hInstance, "Test");
        wc.hCursor        = LoadCursor(NULL, IDC_ARROW);
        wc.hbrBackground  = COLOR_WINDOW + 1;
		wc.lpszMenuName   = "Test";
		wc.lpszClassName  = "Test";


        if (!RegisterClass(&wc))
            return FALSE;
        }


	hLib = LoadLibrary("picbdll.dll");

	if (!hLib)
		return FALSE;


	hWnd = CreateDialog( hInstance, "Test",
						 0, NULL );


    ShowWindow(hWnd, nCmdShow);
    UpdateWindow(hWnd);


    while (GetMessage(&msg, NULL, 0, 0))
        {
        TranslateMessage(&msg);
        DispatchMessage(&msg);
        }

	FreeLibrary( hLib );

    return msg.wParam;
    }



/*
 * TestWndProc
 *
 * Purpose:
 *  Window class procedure.  Standard callback.
 *
 * Parameters:
 *  The standard.  See Section 2.4 Windows SDK Guide to Programming,
 *  page 2-4.
 *
 * Return Value:
 *  See Parameters, above.
 *
 */

long FAR PASCAL TestWndProc(HWND hWnd, UINT iMessage,
							WPARAM wParam, LPARAM lParam)
    {


    switch (iMessage)
        {
        case WM_DESTROY:
            PostQuitMessage(0);
            break;

        case WM_COMMAND:
            switch (wParam)
                {
                case IDM_EXIT:
                    PostMessage(hWnd, WM_CLOSE, 0, 0L);
                    break;
                }
            break;


        default:
            return (DefWindowProc(hWnd, iMessage, wParam, lParam));
        }

    return 0L;
    }
