// ---------- video.cpp

#include <dos.h>
#include <string.h>
#include "video.h"
#include "desktop.h"

unsigned video_mode;
unsigned video_page;
unsigned video_address;
unsigned char ScreenHeight;
unsigned char ScreenWidth;
static union REGS regs;

void initvideo(void)
{
	if (isEGA() || isVGA())	{
		regs.x.ax = 0x1003;
		regs.h.bl = 0;
    	int86(VIDEO, &regs, &regs);
	}
	get_videomode();
}

void restorevideo(void)
{
	if (isEGA() || isVGA())	{
		regs.x.ax = 0x1003;
		regs.h.bl = 1;
	    int86(VIDEO, &regs, &regs);
	}
}

// -------- get the video mode and page from BIOS
void videomode(void)
{
    regs.h.ah = 15;
    int86(VIDEO, &regs, &regs);
    video_mode = regs.h.al;
    video_page = regs.x.bx;
    video_page &= 0xff00;
    video_mode &= 0x7f;
}

// -------- read a character of video memory
unsigned int GetVideoChar(int x, int y)
{
    int c;
    hide_mousecursor();
    c = peek(video_address, vad(x,y));
    show_mousecursor();
    return c & 255;
}

// -------- write a character of video memory
void PutVideoChar(int x, int y, int c)
{
    if (x < ScreenWidth && y < ScreenHeight)    {
        hide_mousecursor();
        poke(video_address, vad(x,y), c);
        show_mousecursor();
    }
}

// --------- Write a string to video memory
void WriteVideoString(char *s, int x, int y, int fg, int bg)
{
    if (x < ScreenWidth && y < ScreenHeight)    {
		int len = strlen(s);
		int *ln = new int[len];
		int *cp1 = ln;
		int col = clr(fg,bg) << 8;
		while (*s)	{
			*cp1++ = (*s & 255) | col;
			s++;
		}
		if (x + len >= ScreenWidth)
			len = ScreenWidth - x;
    	hide_mousecursor();
		movedata(FP_SEG(ln), FP_OFF(ln), video_address, vad(x,y), len*2);
    	show_mousecursor();
		delete ln;
	}
}

// --------- get the current video mode
void get_videomode(void)
{
    videomode();
    // ---- Monochrome Display Adaptor or text mode
    if (ismono())
        video_address = 0xb000;
    else
        // ------ Text mode/
        video_address = 0xb800 + video_page;
	ScreenWidth = *(unsigned char far *)( MK_FP(0x40,0x4a) );
	if (isVGA() || isEGA())
		ScreenHeight = *(unsigned char far *)( MK_FP(0x40,0x84) )+1;
	else
		ScreenHeight = 25;
}

/* -- swap a rectangle of video memory with a save buffer -- */
void swapvideo(Rect &rc, char *bf, Bool Hiding, Bool HasShadow, Bool isFrame)
{
	if (rc.Left() >= ScreenWidth || rc.Top() >= ScreenHeight)
		return;
	char far *hd;
	char far *hadr;
    int ht, bytes_row, bytestobuf, bytesfrbuf, bufferwidth;
    unsigned vadr = vad(rc.Left(), rc.Top());
	Bool TopLine = True, Htrimmed = False, Vtrimmed = False;

    ht = rc.Height();

	if (rc.Bottom() >= ScreenHeight)	{
		Vtrimmed = True;
		ht = ScreenHeight - rc.Top();
	}

    bufferwidth = bytes_row = (rc.Right()-rc.Left()+1) * 2;

	if (rc.Right() >= ScreenWidth)	{
		Htrimmed = True;
		bytes_row = (ScreenWidth - rc.Left()) * 2;
	}

	hd = new char [bytes_row];
	hadr = hd;

    hide_mousecursor();

    while (ht--)    {
		bytestobuf = bytesfrbuf = bytes_row;
		if (HasShadow)	{
			if (TopLine)	{
				if (!Htrimmed)	{
					bytestobuf -= 2;
					bytesfrbuf -= 2;
				}
			}
			else if (ht == 0 && !Vtrimmed)	{
				bytestobuf -= 2;
				bytesfrbuf -= 2;
				vadr += 2;
				hadr += 2;
				bf = (char far *)bf + 2;
			}
			if (!TopLine && !Hiding && !Htrimmed)
				bytesfrbuf -= 2;
		}
		Bool doframe = (Bool)(isFrame && !TopLine && ht != 0);
		if (doframe)
			bytestobuf = bytesfrbuf = 2;
		for (int off = 0; ; off += (rc.Width()-1)*2)	{
        	movedata(video_address, vadr+off, FP_SEG(hadr),
                	FP_OFF(hadr)+off, bytestobuf);
			if (ht || Hiding || !HasShadow || Vtrimmed)
	        	movedata(FP_SEG(bf), FP_OFF(bf)+off, video_address,
                	vadr+off, bytesfrbuf);
        	movedata(FP_SEG(hadr), FP_OFF(hadr)+off, FP_SEG(bf),
                	FP_OFF(bf)+off, bytestobuf);
			if (!doframe || off || Htrimmed)
				break;
		}
        vadr += ScreenWidth*2;
        bf = (char far *)bf + bufferwidth;
		TopLine = False;
    }
	delete hd;
    show_mousecursor();
}

/* -- read a rectangle of video memory into a save buffer -- */
void getvideo(Rect &rc, char *bf)
{
	if (rc.Left() >= ScreenWidth || rc.Top() >= ScreenHeight)
		return;
    int ht = rc.Bottom()-rc.Top()+1;
    int bytes_row = (rc.Right()-rc.Left()+1) * 2;
    unsigned vadr = vad(rc.Left(), rc.Top());
    hide_mousecursor();
    while (ht--)    {
        movedata(video_address, vadr, FP_SEG(bf),
                FP_OFF(bf), bytes_row);
        vadr += ScreenWidth*2;
        bf = (char far *)bf + bytes_row;
    }
    show_mousecursor();
}

/* -- write a rectangle of video memory from a save buffer -- */
void storevideo(Rect &rc, char *bf)
{
	if (rc.Left() >= ScreenWidth || rc.Top() >= ScreenHeight)
		return;
    int ht = rc.Bottom()-rc.Top()+1;
    int bytes_row = (rc.Right()-rc.Left()+1) * 2;
    unsigned vadr = vad(rc.Left(), rc.Top());
    hide_mousecursor();
    while (ht--)    {
        movedata(FP_SEG(bf), FP_OFF(bf), video_address,
                vadr, bytes_row);
        vadr += ScreenWidth*2;
        bf += bytes_row;
    }
    show_mousecursor();
}

/* ---- test for EGA -------- */
Bool isEGA(void)
{
    if (isVGA())
        return False;
    regs.h.ah = 0x12;
    regs.h.bl = 0x10;
    int86(VIDEO, &regs, &regs);
    return (Bool) (regs.h.bl != 0x10);
}

/* ---- test for VGA -------- */
Bool isVGA(void)
{
    regs.x.ax = 0x1a00;
    int86(VIDEO, &regs, &regs);
    return (Bool) (regs.h.al == 0x1a && regs.h.bl > 6);
}

static void Scan350(void)
{
    regs.x.ax = 0x1201;
    regs.h.bl = 0x30;
    int86(VIDEO, &regs, &regs);
}

static void Scan400(void)
{
    regs.x.ax = 0x1202;
    regs.h.bl = 0x30;
    int86(VIDEO, &regs, &regs);
}

/* ---------- set 25 line mode ------- */
void Set25(void)
{
    if (isVGA())	{
        Scan400();
		regs.x.ax = 0x1114;
	}
	else
		regs.x.ax = 0x1111;
    regs.x.bx = 0;
    int86(VIDEO, &regs, &regs);
}

/* ---------- set 43 line mode ------- */
void Set43(void)
{
    if (isVGA())
        Scan350();
    regs.x.ax = 0x1112;
    regs.x.bx = 0;
    int86(VIDEO, &regs, &regs);
}

/* ---------- set 50 line mode ------- */
void Set50(void)
{
    if (isVGA())
        Scan400();
    regs.x.ax = 0x1112;
    regs.x.bx = 0;
    int86(VIDEO, &regs, &regs);
}

/* --------- scroll the screen d: 1 = up, 0 = dn ---------- */
void scroll_window(Rect &rc, int d, int fg, int bg)
{
    hide_mousecursor();
	regs.h.cl = rc.Left();
	regs.h.ch = rc.Top();
	regs.h.dl = rc.Right();
	regs.h.dh = rc.Bottom();
	regs.h.bh = clr(fg,bg);
	regs.h.ah = 7 - d;
	regs.h.al = 1;
    int86(VIDEO, &regs, &regs);
    show_mousecursor();
}


